/* -----------------------------------------

   Create deltaC x-tables for inclusive Higgs production in gluon fusion

   ----------------------------------------- */
#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/time.h>
#include <getopt.h>

#include "hell.hh"
#include "hHiggs_new_var.hh"


using namespace std;


string sas(double as) {
  ostringstream os;
  int ias = int(1000*as+0.1);
  if(1000*as-ias>0.5) ias += 1;
  if     (ias<10  ) os << "000" << ias;
  else if(ias<100 ) os << "00"  << ias;
  else if(ias<1000) os << "0"   << ias;
  else os << ias;
  return os.str();
}


void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -a  --alphas <as>      set alpha_s" << endl
       << " -m  --mH <mH>          set the Higgs mass" << endl
       << " -F  --muF <muF/mH>     set muF/mH" << endl
       << " -R  --RCmode           use the RC variant r->as*b0 (for estimating the uncertainty)" << endl
       << " -v  --variation        use the g+ variant \"Airy\" (for estimating the uncertainty)" << endl
       << " -r  --redo             do NOT recompute the resummed part with 2 M's" << endl
       << " -L  --LLp              use LLp (default is NLL)" << endl
       << endl;
  exit(0);
  return;
}

void read_arguments(int argc, char* argv[], double &as, double &mH, double &muFrat, int &mode, bool &redo, bool &useLLp) {
  const char* const short_options = "ha:m:F:RvrL";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "alphas", 1, NULL, 'a' },
					 { "mH", 1, NULL, 'm' },
					 { "muF", 1, NULL, 'F' },
					 { "RCmode", 0, NULL, 'R' },
					 { "variation", 0, NULL, 'v' },
					 { "redo", 0, NULL, 'r' },
					 { "LLp", 0, NULL, 'L' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'a':
      as = strtod(optarg, NULL);
      break;
    case 'm':
      mH = strtod(optarg, NULL);
      break;
    case 'F':
      muFrat = strtod(optarg, NULL);
      break;
    case 'v':
      mode = 1;
      break;
    case 'R':
      mode = 2;
      break;
    case 'r':
      redo = false;
      break;
    case 'L':
      useLLp = true;
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  return;
}



int main (int argc, char* argv[]) {

  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

  double mH = 125;
  double mt = 173;
  double mb = -4.6;
  double mc = -1.5;
  double as0 = 0.11;
  int nf = 5;
  int localRCmode = 0;
  double muFrat = 1;
  bool forceredo = true;
  bool useLLp = false;
  //
  read_arguments(argc, argv, as0, mH, muFrat, localRCmode, forceredo, useLLp);

  HELL::SetLLpMode(useLLp);
  HELL::LogOrder ord = HELL::NLL;

  HELL::HiggsSmallX *ggHsx = new HELL::HiggsSmallX(mH, mt, mb, mc, muFrat);
  std::vector<std::vector<double> > hk = ggHsx->hCoeffs(3);
  //double c10 = ggHsx->h1()/2;
  double c10 = hk[1][0];
  double c20 = hk[2][0];
  double c11 = hk[2][1];
  double c30 = hk[3][0];
  double c21 = hk[3][1];


  // x grid
  const int Np1 = 300, Np2 = 20-1;
  double x;
  double x_min = 1e-9;
  double x_mid = 0.1;
  double x_max = 0.99+0.01;

  cout << "nf = " << nf << endl;
  HELL::HELLnf sxD(nf, ord, "./data/");
  sxD.SetLargexDamping();
  sxD.SetMomentumConservation(); // has no effect
  sxD.SetEvaluationMethod(1);
  sxD.SetFactorizationScale(muFrat);
  if(localRCmode) sxD.SetRCmode(1);
  vector<double> as1;
  sxD.GetAvailableAlphas(as1);
  // alphas grid
  //double as_min = as1[0];
  //double as_max = as1[as1.size()-1];
  //if(as2[0]>as_min) as_min = as2[0];
  //if(as2[as2.size()-1]<as_max) as_max = as2[as2.size()-1];
  //int Nas = (as_max-as_min)*1000 +1;
  //Nas = (as_max-as_min)*100 +1; // smaller set; comment out for a (10 times) larger set
  //vector<double> asvals;
  //asvals.push_back(as_min);
  //while(asvals[asvals.size()-1]<as_max) {
    //asvals.push_back(asvals[i]+0.001);
    //asvals.push_back(asvals[asvals.size()-1]+0.01);
  //}
  //
  //for(int a=0; a<Nas; a++) {
  //double as = as_min + a*(as_max-as_min)/(Nas-1.);
  for(unsigned int a=0; a<as1.size(); a++) {
    double as = as1[a];
    if(as0>0 && fabs(as-as0)>1e-6) continue;
    if(as0>0) cout << "as = " << as << endl;
    as = int(1000*as+0.1)/1000.;
    //cout << setprecision(10) << 1000*as << "  " << sas(as) << endl;
    ostringstream filename;
    filename << "./data/x-space" << (useLLp ? "-LLp" : "") << "/xtableChiggs";
    if(localRCmode) filename << "var" << localRCmode;
    filename << "_muF" << 1000*muFrat << "_NLL_nf" << nf << "_alphas" << sas(as) << ".table";
    //
    bool infilegood = false;
    double dCx[Np1+Np2], dummy, np1, np2, xmin, xmid, xmax, ic10, ic20, ic11, ic30, ic21;
    ifstream ifile(filename.str().c_str());
    if(ifile.good()) {
      ifile >>  np1 >>  np2 >> xmin >> xmid >> xmax;
      ifile >> ic10 >> ic20 >> ic11 >> ic30 >> ic21;
      if(np1==Np1 && np2==Np2 && xmin==x_min && xmid==x_mid && xmax==x_max) {
	//if(ic10==c10 && ic20==c20 && ic11==c11 && ic30==c30 && ic21==c21) {
	for(int i=0; i<Np1+Np2; i++) {
	  ifile >> dCx[i] >> dummy;
	}
	infilegood = true;
      }
      ifile.close();
    }
    ofstream ofile(filename.str().c_str());
    ofile << VERSIONx << endl;
    ofile
      << setw(15) << Np1
      << setw(15) << Np2
      << setw(15) << x_min
      << setw(15) << x_mid
      << setw(15) << x_max
      << endl;
    ofile << setw(15) << c10
	  << setw(15) << c20
	  << setw(15) << c11
	  << setw(15) << c30
	  << setw(15) << c21
	  << endl;
    //
    double dC, dCaux;
    for(int i=0; i<Np1+Np2; i++) {
      if(i<Np1) x = x_min * exp(i/(Np1-1.)*log(x_mid/x_min));
      else      x = x_mid + (i-Np1+1)*(x_max-x_mid)/(Np2-0.);
      //cout << x << endl;
      dCaux = x*sxD.deltaHggAux(as, x, mH, mt, mb, mc);
      //cout << "bravo" << endl;
      if(infilegood && !forceredo) {
	dC = dCx[i];
      } else {
	dC = x*sxD.deltaHgg(as, x, mH, mt, mb, mc);
      }
      ofile
	//<< setw(15) << x
	<< setw(15) << dC
	<< setw(15) << dCaux
	<< endl;
      cout << "-";
      cout.flush();
    }
    cout << ".";
    cout.flush();
    //
    ofile.close();
  }
  cout << endl;


  // Finish time
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << "Total time: " << t << "s" << endl;

  return 0;

}
