#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <cmath>
#include <cstdlib>
//#include <cuba.h>
#include <gsl/gsl_errno.h>

#include "CoefficientFunctionsNew.hh"
#include "CoefficientFunctionsOffShell.hh"
#include "CoefficientFunctionsOffShellInterpolator.hh"
#include "hHiggs_new_var.hh"
#include "Pqg.hh"
#include "math/borel.hh"
#include "math/special_functions.hh"
#include "integration.hh"
#include "version.hh"
#include "h_coefficients.hh"
//#include <complex_bessel/sp_airy.h>


using namespace std;



namespace HELL {


  unsigned int RCmodeNew = 0;
  void SetRCmodeCFnew(unsigned int mode) {  // mode=0 uses gammadot, mode=1 uses as*beta0
    if(mode>1) {
      cout << "Error: invalid RC mode. Using default." << endl;
      mode = 0;
    }
    RCmodeNew = mode;
  }
  bool RCvarNew = 0;
  void SetRCvarCFnew(bool var) {  // var=true uses "Aiy" anomalous dimension
    RCvarNew = var;
  }

  xTableU::xTableU(string filename) {
    ifstream infile(filename.c_str());
    if(!infile.good()) {
      cout << "\033[0;31m" << "HELLx: Error reading table " << filename << "\033[0m" << endl;
      exit(0);
    }
    string vers;
    infile >> vers >> Np1 >> Np2 >> x_min >> x_mid >> x_max >> Nxi;
    if(vers!=VERSION) {
      std::cout << "\033[0;31m" << "HELL: Error! The U tables you are trying to read are version " << vers << " which is not compatible with the code version " << VERSION << "\033[0m" << std::endl;
      abort();
    }
    Nxi--; // this patches a bad choice in the produced files
    xx    = new double[Np1+Np2];
    xdUqg = new double[Np1+Np2];    
    for(int i=0; i<Np1+Np2; i++) {
      if(i<Np1) xx[i] = x_min * exp(i/(Np1-1.)*log(x_mid/x_min));
      else      xx[i] = x_mid + (i-Np1+1)*(x_max-x_mid)/(Np2-0.);
      infile >> xdUqg[i];
    }
    //}
    //void xTableU::Init() {
    xi_ = new double [Nxi];
    xdU = new double*[Nxi];
    for(int m=0; m<Nxi; m++) {
      xdU[Nxi-1-m] = new double[Np1+Np2];
      infile >> xi_[Nxi-1-m];
      for(int i=0; i<Np1+Np2; i++) {
	infile >> xdU[Nxi-1-m][i];
      }
      //cout << Nxi-1-m << "   "<< xi_[Nxi-1-m] << "  " << xdU[Nxi-1-m][0] << "  " << xdU[Nxi-1-m][Np1+Np2-1] << endl;
    }
    infile.close();
    dUspline = new gsl_spline;
    dUspline = gsl_spline_alloc (gsl_interp_cspline, Nxi);
  }
  double xTableU::interpolate(double x) {
    if(x>1 || x<0) {
      cout << "\033[0;31m" << "HELLx: Error! Requesting resummed splitting function for unphysical value of x=" << x << " outside the physical range 0<x<=1" << "\033[0m" << endl;
      exit(45);
    }
    if(x>x_max) {
      cout << "\033[0;31m" << "HELLx: Warning! Extrapolating out of interpolation range: x=" << x << " > x_max=" << x_max << "\033[0m" << endl;
      x = x_max;
    }
    if(x<x_min) {
      cout << "\033[0;31m" << "HELLx: Warning! Extrapolating out of interpolation range: x=" << x << " < x_min=" << x_min << "\033[0m" << endl;
      x = x_min;
    }
    double ii;
    if(x<x_mid) ii = (Np1-1.)*log(x/x_min)/log(x_mid/x_min);
    else        ii = Np1-1.+Np2*(x-x_mid)/(x_max-x_mid);
    return ii;
  }
  double xTableU::evalU(double x, double xi, double &deriv) {
    double ii = interpolate(x);
    int i = int(ii);
    if(i<0 || ii<i) {
      cout << "\033[0;31m" << "HELLx: Error! This should never happen" << "\033[0m" << endl;
      abort();
    }
    double *xidU;
    xidU = new double[Nxi];
    for(int m=0; m<Nxi; m++) {
      xidU[m] = xdU[m][i] + (i==Np1+Np2-1 ? 0 : (ii-i)*(xdU[m][i+1]-xdU[m][i]));
    }
    // spline setup
    double dU = 0;
    deriv = 0;
    if     (xi>xi_[Nxi-1]) dU = xidU[Nxi-1];
    else if(xi<xi_[0])     dU = xidU[0];
    else {
      gsl_spline_init (dUspline, xi_, xidU, Nxi);
      gsl_interp_accel *acc = gsl_interp_accel_alloc ();
      dU    = gsl_spline_eval      (dUspline, xi, acc);
      deriv = gsl_spline_eval_deriv(dUspline, xi, acc);
      gsl_interp_accel_free (acc);
      //gsl_spline_free(dUspline);
    }
    delete[] xidU;
    return dU;
  }
  double xTableU::evalUqg(double x) {
    double ii = interpolate(x);
    int i = int(ii);
    if(i<0 || ii<i) {
      cout << "\033[0;31m" << "HELLx: Error! This should never happen" << "\033[0m" << endl;
      abort();
    }
    double dU;
    dU = xdUqg[i] + (i==Np1+Np2-1 ? 0 : (ii-i)*(xdUqg[i+1]-xdUqg[i]));
    return dU;
  }






  double integrand(double xi, double x, xTableU *tU, process p, double asb0, double muFrat, void *par) {
    double dU = 0;
    double xiU = xi/muFrat/muFrat;  // use the variable ( Q^2/muF^2 * xi )
    double  U = tU->evalU(x, xiU, dU);
    if(xiU<1) {  // apply higher-twist damping
      int kk = 1;
      double  HTdamp = 1 - pow(-asb0*log(xiU),kk*(1+1/asb0));
      double dHTdamp = (HTdamp-1) * kk*(1+1/asb0)/xiU/log(xiU);
      dU = dU * HTdamp + U * dHTdamp;
      U *= HTdamp;
    }
    //
    //cout << x << "  " << xi << endl;
    double res = U;
    //
    if(p == gamma_qg) res *= - Dxi_Hqg(xi);
    //
    //if(p == CL) res *= - Dxi_CLoff(xi, x1);
    //if(p == C2) res *= - Dxi_C2off(xi, x1) - (xi<=1 ? 1./xi : 0);
    if(p == CL) res *= - Dxi_CLoff(xi);
    if(p == C2) res *= - Dxi_C2off(xi) - (xi<=1 ? 1./xi : 0);
    //if(p == CL) res = CLoff(xi) * dU;
    //if(p == C2) res = (C2off(xi) + (xi<=1 ? log(xi) : 0)) * Dxi_evolU(xi,gamma,r);
    //
    //if(p == Ctest1) res = - (xi<=1 ? 1. : 0) * U + (xi<=1 && xi>=0.9 ? 10. : 0);
    //if(p == Ctest2) res = (xi<=1 ? xi : 0) * Dxi_evolU(xi,gamma,r);
    //if(p == Ctest1) res = - Dxi_CLoff(xi) * U;
    //if(p == Ctest2) res = CLoff(xi) * Dxi_evolU(xi,gamma,r);
    //
    if(p>=50 && p<60) {
      double mQ = *(double*) par;
      if(p == Khg) res *= - Dxi_Khgoff(xi,mQ) + (xi<=1 ? 1./xi : 0);
    }
    if(p >= 60 && p<100) {
      Interpolator *interp = (Interpolator*) par;
      if(p == MC2)   res *= - Dxi_MC2off(xi, interp);
      if(p == MCL)   res *= - Dxi_MCLoff(xi, interp);
      if(p == MC2ax) res *= - Dxi_MC2axoff(xi, interp);
      if(p == MCLax) res *= - Dxi_MCLaxoff(xi, interp);
      //
      double xim = interp->Getxim();
      if(p == MC2cc) res *= - Dxi_MC2CCoff(xi, interp) -               (xi<=1 ? 1./xi : 0) /2;
      if(p == MCLcc) res *= - Dxi_MCLCCoff(xi, interp) - xim/(1+xim) * (xi<=1 ? 1./xi : 0) /2;
      if(p == MC3cc) res *= - Dxi_MC3CCoff(xi, interp) -   1/(1+xim) * (xi<=1 ? 1./xi : 0) /2;
    }
    if(p == ggHaux) {
      HiggsSmallX *ggHsx = (HiggsSmallX*) par;
      res *= ggHsx->sf1y1dt(xi/2)/2;
    }
    if(p == DY) {
      res *= - Dxi_CDYoff(xi) - (xi<=1 ? 1./xi : 0);
    }
    //
    //cout << setw(15) << xi << setw(15) << x1 << setw(35) << res << endl;
    return res;
  }
  struct s_prdtv { process p; double d1, d2, d3, d4; xTableU *tU; void *par; };
  double integrandGSLx(double x, void *pars) {
    s_prdtv p = *(s_prdtv*)pars;
    double xx  = p.d1;
    //xTableU *tU = p.tU;
    double a0 = p.d2;  // as(Q^2)
    double b0 = p.d3;  // beta0
    double muFrat = p.d4;
    double lFQ = 2*log(muFrat);
    //double xi0 = real(r); // may be negative - problematic
    //double xi0 = a0*b0;
    //if(real(r)>0 && real(r)>xi0) xi0 = real(r); // not sure if it makes sense...
    //double xi = exp(-1./a0/b0) + (1.-x)/x;  double jac = 1./x/x;
    //double mult = 3; double t = (mult*x-1)/a0/b0; double xi = exp(t);  double jac = xi*mult/a0/b0;
    //double mult = 3; double u0=1./a0/b0-lFQ; double t = x*u0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1*u0; double jac2 = mult*xi2*u0;
    double mult = 3; if(a0<0.007) mult=2; double u01=1./a0/b0-lFQ; double t1 = x*u01; double u02=1./a0/b0; double t2 = x*u02; double xi1 = exp(-t1); double xi2 = exp(mult*t2); double jac1 = xi1*u01; double jac2 = mult*xi2*u02;
    //double mult = 3; double u=x*x; double t = u/a0/b0; double xi1 = exp(-t); double xi2 = exp(mult*t); double jac1 = xi1/a0/b0*2*x; double jac2 = mult*xi2/a0/b0*2*x;
    //double amin = 0.01*a0, amax = 2000*a0;  double a = amin + (amax-amin)*x;  double xi = exp((1/a-1/a0)/b0);  double jac = -xi/(-b0*a*a) * (amax-amin);
    //double Xmin = 0.1, Xmax = 80;  double X = Xmin + (Xmax-Xmin)*x;  double xi = exp((1-X)/X/xi0);  double jac = xi/X/X/xi0*(Xmax-Xmin);
    //double Xmin = 0.05, Xmax = 70;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin); double X2 = 1 + (Xmax-1)*(1-x);  double xi2 = exp((1-X2)/X2/xi0);  double jac2 = xi2/X2/X2/xi0*(Xmax-1);
    //double Xmin = 0.1;  double X1 = Xmin + (1-Xmin)*x;  double xi1 = exp((1-X1)/X1/xi0);  double jac1 = xi1/X1/X1/xi0*(1-Xmin);  double X2 = x;  double xi2 = exp((1-1/X2)*X2/xi0);  double jac2 = xi2/xi0;
    //cout << x << "  " << xi << endl;
    //cout << xi1 << "   " << xi2 << endl;
    //cout << u01 << "  " << lFQ << endl;
    //if(1+real(r)*log(xi1)<0) cout << "warning! xi1="<< xi1 << "  r="<< r << endl;
    //if(1+real(r)*log(xi2)<0) cout << "warning! xi2="<< xi2 << "  r="<< r << endl;
    //
    //cout << xi0 << "  " << a0 << "  "  << b0 << "  " << xi2 << "  " << x << endl;
    //if(x>1e-5) cout << xi0 << "  " << a0 << "  "  << b0 << "  " << xi << "  " << x << endl;
    //double result = jac  * integrand(xi,  xx, p.tU, p.p, a0*b0, muFrat, p.par);
    double result = jac1 * integrand(xi1, xx, p.tU, p.p, a0*b0, muFrat, p.par) + jac2 * integrand(xi2, xx, p.tU, p.p, a0*b0, muFrat, p.par);
    //if(result==result) cout << result << endl;
    //
    //cout << setw(15) << xi << setw(15) << x[1] << setw(35) << res << endl;
    //delete tU;
    return result;
  }


  extern bool a10woRC, fullySymm, shiftMmin;
  extern double a11();
  extern double a10(int nf);
  extern double a21(int nf);
  extern double a20(int nf);
  extern double a32(int nf);
  const double c0 = CA/M_PI*4*log(2.);
  const double k0 = CA/M_PI*28.*ZETA3;
  // from DL-LO kernel
  inline double c1LO (int nf) { return -15.00496429 - 0.04503163717*nf;}
  inline double k1LO (int nf) { return -507.744719  - 1.080759292*nf;}
  // from DL-NLO kernel
  //inline double c1NLO(int nf) { return -15.04043249 - 0.2076545137*nf; }
  //inline double k1NLO(int nf) { return -574.496771  - 0.3724353405*nf; }
  //inline double m1NLO(int nf) { return 0.3099881585 - 0.01878716112*nf; }
  inline double c1NLO(int nf) { return -11.696833425 - 0.4102968810*nf; }
  inline double k1NLO(int nf) { return -494.250393369 - 5.23585215538*nf; }
  inline double m1NLO(int nf) { return -0.1492429211 + 0.00904502552*nf; }
  //
  unsigned int factorials[] = {1,1,2,6,24,120,720};
  unsigned int factorial(unsigned int k) {
    if(k<7) return factorials[k];
    double res = 1;
    for(int j=1; j<=k; j++) res *= j;
    return res;
  }
  double Pole(double x, int k, int j) { // inverse Mellin of 1/N^k/(1+N)^j
    if(j==0 && k>0) return pow(-log(x),k-1)/factorial(k-1)/x;
    if(k==0 && j>0) return pow(-log(x),j-1)/factorial(j-1);
    if(k==1 && j==1) return 1/x-1;
    if(k==2 && j==1) return Pole(x,2,0) - Pole(x,1,0) + Pole(x,0,1);
    if(k==1 && j==2) return Pole(x,1,0) - Pole(x,0,2) - Pole(x,0,1);
    cout << "HELLx warning: this inverse mellin is not implemented: 1/N^"<<k<<"/(1+N)^"<<j << endl;
    return 0;
  }
  double exactPoly(double x, int k, int j) { // inverse Mellin of 1/N^k/(1+N)^j (psi_1(1+N)-Zeta2)
    double res = 0;
    if     (k==2 && j==0) res = (2*(Li3(x)-ZETA3) - Li2(x)*log(x))/x;
    else if(k==3 && j==0) res = (3*(ZETA4-Li4(x))+(Li3(x)+2*ZETA3)*log(x))/x;
    else if(k==0 && j==2) res = 2*(Li3(x)-ZETA3) - Li2(x)*log(x) - pow(log(x),3)/6;
    else if(k==0 && j==3) res = 3*(ZETA4-Li4(x))+(Li3(x)+2*ZETA3)*log(x) + pow(log(x),4)/24;
    else cout << "HELLx warning: this inverse mellin is not implemented: (psi_1(1+N)-Zeta2)/N^"<<k<<"/(1+N)^"<<j << endl;
    return res;
  }
  //
  double P0(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = a1/x + a0;
    return res;
  }
  double P0sq(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = -a1*a1*log(x)/x - a0*a0*log(x) + 2*a1*a0*(1/x-1);
    return res;
  }
  double P0cb(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = a1*a1*a1*Pole(x,3,0) + a0*a0*a0*Pole(x,0,3) + 3*a1*a1*a0*Pole(x,2,1) + 3*a1*a0*a0*Pole(x,1,2);
    return res;
  }
  double P1(double x, int nf) {
    double res = 0;
    if(useLLp) {
      res = beta0(nf)*(3*k0/32.-c0) * (1/x-4*(1+log(x)));
    } else {
      double a1 = a21(nf);
      res = a1 * (1/x-2);
    }
    return res;
  }
  double P0P1(double x, int nf) {
    double a1 = a11(), a0=a10(nf) - (a10woRC ? beta0(nf) : 0 );
    double res = 0;
    if(useLLp) {
      res = beta0(nf)*(3*k0/32.-c0) * ( a1*(Pole(x,2,0)-4*Pole(x,0,2)) + a0*(Pole(x,1,0)+4*Pole(x,0,3)-4*Pole(x,0,2)-Pole(x,0,1)) );
    } else {
      res = a21(nf) * ( a1*Pole(x,2,0) + (a0-2*a1)/x -2*a0*Pole(x,0,2) + 2*a1-a0 );
    }
    return res;
  }
  double P2(double x, int nf, int var) {
    double res = 0;
    double a1 = a11(), a0=a10(nf);
    double chi02 = a1*a0;
    double b0 = beta0(nf);
    if(useLLp) {
      res = a1*chi02 * ( Pole(x,2,0) - 3*Pole(x,1,0) + 2*Pole(x,0,2) + 3*Pole(x,0,1) )
	+ 2*(a0 + a1)*chi02 * (2*Pole(x,0,4)-Pole(x,0,3));
      res += CA/M_PI* ( a1*a1*  exactPoly(x,3,0) 
			+a1*a0* exactPoly(x,2,0)
			+4*pow(a0 + a1,2)*              exactPoly(x,0,3)
			-2*(a0*a0 + 4*a0*a1 + 3*a1*a1)* exactPoly(x,0,2)
			);
      // add RC
      int T = (var==1 ? 1 : 2);
      double  c1 = c1NLO(nf);
      double  k1 = k1NLO(nf);
      double  m1 = m1NLO(nf);
      double dc1 = c1-c1LO(nf);
      double dk1 = k1-k1LO(nf);
      double dm1 = m1;
      if(shiftMmin) m1 += c0/2.;
      double lam2 = b0*(48*c0*k0-3*k0*k0-256*c0*c0)/256. + (16*c0*(dk1+6*k0*dm1)+k0*(16*dc1-3*dk1-15*k0*dm1))/512.;
      double lam1 = b0*(3*T*k1/32 -T*c1 + (b0+6*m1)*k0/16.);
      res += lam1*Pole(x,1,0) + lam2*Pole(x,2,0) - 4*(1+log(x))*(lam1+lam2);
    } else {
      double chi01 = a1;
      double chi11 = a21(nf);
      double rho = chi01*chi01*(5*ZETA3/2.-1./24.) + chi02*(53./18.-ZETA2) - chi11;
      //double rho = ( CA*CA*(-74./27. +11./12.*ZETA2 +5./2.*ZETA3) +nf*((4*CA+7*CF)/27. +(CA-2*CF)*ZETA2/6.) )/M_PI/M_PI;
      res = (rho*a1 + a1*chi02 + a1*chi11) * Pole(x,2,0)
	+ (-4*rho*a1 - 3*a1*chi02 - 3*a1*chi11) * Pole(x,1,0)
	+ 4*(a0 + a1)*(2*rho + chi02 + chi11) * Pole(x,0,4)
	- 2*(a0 + a1)*(4*rho + chi02 + chi11) * Pole(x,0,3)
	+ 2*(rho*a0 + 3*rho*a1 + a1*chi02 + a1*chi11) * Pole(x,0,2)
	+ (4*rho*a1 + 3*a1*chi02 + 3*a1*chi11) * Pole(x,0,1)
	;
      res += CA/M_PI* ( a1*a1*  exactPoly(x,3,0) 
			+a1*a0* exactPoly(x,2,0)
			+4*pow(a0 + a1,2)*              exactPoly(x,0,3)
			-2*(a0*a0 + 4*a0*a1 + 3*a1*a1)* exactPoly(x,0,2)
			);
      if(!a10woRC && !fullySymm) res += b0 * chi01 * (a1*exactPoly(x,2,0) + 2*(a1+a0)*exactPoly(x,0,2) + a1*4*(Pole(x,1,0)-Pole(x,0,1)-Pole(x,0,2)) +2*(a1+a0)*(Pole(x,0,3)-Pole(x,0,4)) );
      // add RC
      res += b0*b0*k0/16* (1/x-4*(1+log(x)));
    }
    // result
    return res;
  }


  double newDeltaFRes(double x, double as, double nf, xTableU *tU, process proc, int ordersub=0, double muFrat=1, void *pars=NULL) {
    hCoeff hC(proc, pars);
    //double h0 = hC.hk(0);
    double h1 = hC.hk(1);
    double h2 = hC.hk(2);
    double h3 = hC.hk(3);
    //
    double b0 = beta0(nf);
    //
    s_prdtv intpars = { proc, x, as, b0, muFrat, tU, pars };
    gsl_error_handler_t *old_handler = gsl_set_error_handler (NULL);
    gsl_set_error_handler_off();
    double prec = 1e-4;
    int status;
    double res = integration_gauss(integrandGSLx, 0, 1, prec, NULL, &intpars, &status);
    if(status) res = 0;
    gsl_set_error_handler (old_handler);
    //double dU = 0;
    //double  U = tU->eval(x, exp(-1./as/b0), dU);
    //res += U*CLoff(exp(-1./as/b0));
    //
    double xP0 = x*P0(x,nf);
    //double P1 = P1sing(x,nf);
    //if(ordersub>0) res -= h0;  // this is a delta function in z=1 -> ignore it
    if(ordersub>1) res -= h1*as*xP0;
    if(ordersub>2) {
      double xP0sq = x*P0sq(x,nf);
      double xP1   = x*P1(x,nf);
      res -= as*as*(h2*(xP0sq-b0*xP0) + h1*xP1); // for DIS coefficient functions matched to NNLO we don't need this
      if(ordersub>3) {
	double xP0cb = x*P0cb(x,nf);
	double xP0P1 = x*P0P1(x,nf);
	double xP2   = x*P2(x,nf,RCvarNew);
	double T = (RCmodeNew==1 ? 1 : 2);
	res -= as*as*as * (  h3*(xP0cb-3*b0*xP0sq+2*b0*b0*xP0) + h2*(2*xP0P1-T*b0*xP1) + h1*xP2);
      }
    }
    if(ordersub>4) cout << "Error, trying to subtract beyond the capability of HELL" << endl;
    //
    return res;
  }


  double DeltaUqgRes(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    double res = tU->evalUqg(x);
    return res;
  }

  // matching condition
  double DeltaKhgRes(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, Khg, 2, muFrat, pars);
    res *= as/3./M_PI;
    res += tU->evalUqg(x)/nf;
    return res;
  }

  // massless DIS Coeff Functions
  double DeltaCLRes(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, CL, 2, muFrat);
    return res * as*nf/3./M_PI;
  }
  double DeltaC2Res(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, C2, 2, muFrat);
    res *= as*nf/3./M_PI;
    res -= tU->evalUqg(x);
    return res;
  }

  // massive DIS Coeff Functions in Neutral Current
  double DeltaMC2Res(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MC2, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }
  double DeltaMCLRes(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MCL, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }
  // axial contributions
  double DeltaMC2ResAxial(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MC2ax, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }
  double DeltaMCLResAxial(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+4*interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MCLax, 2, muFrat, pars);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res * as/3./M_PI;
  }

  // massive DIS Coeff Functions in Charged Current
  double DeltaMC2ResCC(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double X = 1/(1+interp->Getxim());
    if(x>=X) return 0;
    double res = newDeltaFRes(x, as, nf, tU, MC2cc, 2, muFrat, pars);
    //res *= 1-pow(x/X,1./(1-X));  // move below? if not, it has to be applied before subtraction of the singularity, in integrand()
    res *= as/3./M_PI;
    res -= tU->evalUqg(x)/nf/2.;
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res;
  }
  double DeltaMCLResCC(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double xim = interp->Getxim();
    double X = 1/(1+xim);
    if(x>=X) return 0;    
    double res = newDeltaFRes(x, as, nf, tU, MCLcc, 2, muFrat, pars);
    //res *= 1-pow(x/X,1./(1-X));  // move below?
    res *= as/3./M_PI;
    res -= tU->evalUqg(x)/nf/2. * xim/(1+xim);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res;
  }
  double DeltaMC3ResCC(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    Interpolator *interp = (Interpolator*) pars;
    double xim = interp->Getxim();
    double X = 1/(1+xim);
    if(x>=X) return 0;    
    double res = newDeltaFRes(x, as, nf, tU, MC3cc, 2, muFrat, pars);
    //res *= 1-pow(x/X,1./(1-X));  // move below?
    res *= as/3./M_PI;
    res -= tU->evalUqg(x)/nf/2. /(1+xim);
    res *= 1-pow(x/X,1./(1-X));
    //delete interp;
    return res;
  }

  
  double DeltaHggAuxRes(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, ggHaux, 4, muFrat, pars);
    return res;
  }

  double DeltaCDYRes(double x, double as, double nf, xTableU *tU, double muFrat, void *pars) {
    double res = newDeltaFRes(x, as, nf, tU, DY, 2, muFrat, pars);
    res *= as/6./M_PI;
    res -= tU->evalUqg(x)/nf/2.;
    return res;
  }


};

